#!/usr/local/env python3
import numpy as np
import pandas as pd
import fileinput
import os
try:
	from pyminer.common_functions import read_table, get_sample_k_lists, import_dict
except:
	from common_functions import read_table, get_sample_k_lists, import_dict


def load_relative_table(in_dir, rel_path, out_type="numpy", no_head = False):
	in_file = os.path.join(in_dir,rel_path)
	if os.path.isfile(in_file):
		if out_type == "numpy":
			return(np.array(read_table(in_file)))
		if out_type == "pandas":
			if no_head:
				return(pd.read_csv(in_file, sep='\t', header=None))
			else:
				return(pd.read_csv(in_file, sep='\t'))
		if out_type == "list":
			return(read_table(in_file))
	else:
		print("\n\nWARNING: Could not find the file: "+in_file+"\n\n")
		return(None)

##################

class ap_obj(object):
	def __init__(self, in_dir):
		self.in_dir = in_dir
		self.load_ap_interactions()

	def load_ap_interactions(self):
		self.all_num_clust_interactions = load_relative_table(self.in_dir, "autocrine_paracrine_signaling/all_cell_cell_interaction_summary.tsv", out_type = "pandas")
		self.pathway_neg_log_p_vals = load_relative_table(self.in_dir, "autocrine_paracrine_signaling/combined_neg_log10p_gprofiler.tsv", out_type = "pandas")
		self.pathway_importance = load_relative_table(self.in_dir, "autocrine_paracrine_signaling/individual_class_importance.tsv", out_type = "pandas")
		#autocrine_paracrine_signaling/extracellular_extracellular_cell_type_specific_interactions.txt
		#autocrine_paracrine_signaling/extracellular_plasma_membrane_cell_type_specific_interactions.txt
		#autocrine_paracrine_signaling/plasma_membrane_plasma_membrane_cell_type_specific_interactions.txt

	def load_clust_clust_interactions(self, clust_1, clust_2):
		print("loading all the interactions between", clust_1, "and", clust_2)
		## filter for only the correct ones & return a pandas table
		## clust_1: 1
		## clust_2: 5
		"autocrine_paracrine_signaling/all_cell_type_specific_interactions.tsv"
		return()

	def load_pathway(self, pathway):
		## do a grep like thing for this table and the pathway
		## return a pandas table containing only the lines pertinent to the pathway
		"autocrine_paracrine_signaling/all_cell_type_specific_interactions_gprofiler.tsv"
		return()

	def load_clust_clust_pathways(self, clust_1, clust_2):
		## do a grep like thing for this table and the pathway
		## return a pandas table containing only the lines pertinent to the pathway
		"autocrine_paracrine_signaling/all_cell_type_specific_interactions_gprofiler.tsv"
		return()

	def __str__(self):
		out_str = "PyMINEr autocrine paracrine analyses"
		out_str += "\nObjects of interest:"
		out_str += "\n\t<object>.all_num_clust_interactions: An adjacency list of all clusters with all other \n\t\tclusters and the number of interactions that they have with each other."
		out_str += "\n\t<object>.pathway_neg_log_p_vals: Pathway (row) -log10(p-values) and their corresponding \n\t\tclusters-cluster pairs & their signaling interactions (columns). \n\t\tTable is sorted by how informative a given pathway is."
		out_str += "\n\t<object>.pathway_importance: A 2D table of pathways (rows) & their importance for each \n\t\tcluster-cluster pair (column). Sort this table by a given column \n\t\tto get the most informative pathways that describe the given gene module."
		out_str += "\n\n\nFunctions of interest"
		out_str += "\n\t<object>.: "
		return(out_str)

	def __repr__(self):
		return("ap_obj('"+self.in_dir+"')")

##################
class goi_obj(object):
	def __init__(self, in_dir):
		self.in_dir = in_dir
		self.load_goi()

	def load_goi(self):
		self.goi_dir = os.path.join(self.in_dir,"genes_of_interest")
		if os.path.isdir(self.goi_dir):
			goi_short_path_file = os.path.join(self.goi_dir,"genes_of_interest_shortest_path_list.txt")
			if os.path.isfile(goi_short_path_file):
				self.goi_shortest_paths = load_relative_table(self.in_dir, goi_short_path_file, out_type = "pandas")
		else:
			self.goi_dir = None
			self.goi_shortest_paths = None

	def __str__(self):
		out_str = "PyMINEr genes of interest analyses"
		out_str += "\nObjects of interest:"
		out_str += "\n\t<object>.goi_dir: The directory that contains the plots that were automatically \n\t\tgenerated for your genes of interest"
		out_str += "\n\t<object>.goi_shortest_paths: A 2D table that contains your genes of interest (columns) \n\t\tand all other genes, and indicates the shortest path distance in \n\t\tthe co-expression network graph from the pertinent gene of interest."
		return(out_str)

	def __repr__(self):
		return("goi_obj('"+self.in_dir+"')")

##################
class network_obj(object):
	def __init__(self, in_dir):
		print("loading network analysis")
		self.in_dir = in_dir
		self.load_modules()

	def load_modules(self):
		self.communities = load_relative_table(self.in_dir, "pos_cor_graphs/communities.txt", out_type = "pandas")
		self.community_cluster_use_stats = load_relative_table(self.in_dir, "pos_cor_graphs/community_analysis/global_statistics.tsv", out_type = "pandas")
		self.pathway_neg_log_p_vals = load_relative_table(self.in_dir, "pos_cor_graphs/community_analysis/combined_neg_log10p_gprofiler.tsv", out_type = "pandas")
		self.pathway_importance = load_relative_table(self.in_dir, "pos_cor_graphs/community_analysis/individual_class_importance.tsv", out_type = "pandas")

	def load_individual_modules(self):
		pass
		# module_dict = {}
		# ## 
		# for 

	def load_network(self, full_network=False):
		## this is not run automatically because these networks can be large
		self.graph_full_file = os.path.join(self.in_dir,"pos_cor_graphs/full_graph_full.graphpkl")
		self.graph_file = os.path.join(self.in_dir,"pos_cor_graphs/large_comps.graphpkl")
		pass

	def __str__(self):
		out_str = "PyMINEr network analyses"
		out_str += "\nObjects of interest:"
		out_str += "\n\t<object>.communities: Annotations of all genes and their co-expression module/community,\n\t\tas well as some other stats like page rank and local page rank (the later is an \n\t\talgorithm not yet published)"
		out_str += "\n\t<object>.community_cluster_use_stats: Statistics for whether a module is differentially\n\t\tutilized across clusters."
		out_str += "\n\t<object>.pathway_neg_log_p_vals: Pathway (row) -log10(p-values) and their corresponding\n\t\tclusters (columns). Table is sorted by how informative a given pathway is."
		out_str += "\n\t<object>.pathway_importance: A 2D table of pathways (rows) & their importance for each \n\t\tgene module/community (column). Sort this table by a given column to get the most \n\t\tinformative pathways that describe the given gene module."
		# out_str += "\n\t<object>.: "
		# out_str += "\n\t<object>.: "
		return(out_str)

	def __repr__(self):
		return("network_obj('"+self.in_dir+"')")

##################
class clust_plotting_obj(object):
	def __init__(self, in_dir):
		self.in_dir = in_dir
		self.load_cluster_plotting()

	def load_cluster_plotting(self):
		self.cluster_pkl = import_dict(os.path.join(self.in_dir,"sample_clustering_and_summary/clustering_plots.pkl"))
		## all plots
		self.plotting_subset = self.cluster_pkl["sample_indices_for_plotting"]
		self.cluster_colors = self.cluster_pkl["color_vect"]
		self.linear_groups = self.cluster_pkl["linear_groups"]
		self.sample_k_lists = self.cluster_pkl["sample_k_lists"]
		## 2D plots
		self.plots = self.cluster_pkl["plots"]
		## heatmap
		self.group_reordering_vector = self.cluster_pkl["group_reordering_vector"]
		self.reordered_colors = self.cluster_pkl["reordered_colors"]
		## cell embedding
		cell_embedding_file = os.path.join(self.in_dir,"sample_clustering_and_summary/cell_embedding.graphpkl")
		if os.path.isfile(cell_embedding_file):
			self.cell_embedding_network = import_dict(cell_embedding_file)
		else:
			self.cell_embedding_network = "clustering wasn't done with Louvain modularity, so there's no graph network representation of the cell embedding"

	def __str__(self):
		out_str = "\nPyMINEr cluster plotting analyses"
		out_str += "\nObjects of interest:"
		out_str += "\n\n:Objects relevant to all plots"
		out_str += "\n\t<object>.plotting_subset: These are really important: only a subset of cells are \n\t\ttypically use for plotting in datasets of sufficient size (default=15k). This vector are the \n\t\tindices of that subset from the original dataset."
		out_str += "\n\t<object>.cluster_colors: The vector of colors used by PyMINEr to make all of the \n\t\tcluster-colorized plots."
		out_str += "\n\t<object>.linear_groups: The group level annotations for the subset of cells used."
		out_str += "\n\t<object>.sample_k_lists: A re-organization of linear_groups that is a list of lists, \n\t\twith the first level being the cluster number, and each of those lists being the indices \n\t\tof the cells that belong to that cluster. Note again that the indices are for the \n\t\tsubset of cells used for plotting."
		out_str += "\n\n2D-scatter related objects:"
		out_str += "\n\t<object>.plots: 2D projections of the <object>.plotting_subset subset of cells & \n\t\ttheir x,y coords, color,and the file name for where that image is saved. This could be useful \n\t\tif you want to use the same UMAP x/y projection for additional plots or something like that."
		out_str += "\n\nHeatmap related objects:"
		out_str += "\n\t<object>.group_reordering_vector: After subsetting the expression matrix for the \n\t\t<object>.plotting_subset cells, use this vector to reorder that subset to plot a heatmap that \n\t\tgoes in serial from cluster 0 forward."
		out_str += "\n\t<object>.reordered_colors: the vector of colors that has been re-ordered to go in \n\t\tserial from cluster 0 onward"
		out_str += "\n\nNetwork graph embedding of ALL cells (Only if you used Louvain modularity as the \n\t\tclustering method)"
		out_str += "\n\t<object>.cell_embedding_network: the networkx graph object containing the network \n\t\tof the pertinent subset of cells."
		# out_str += "\n\t<object>.:"
		return(out_str)

	def __repr__(self):
		return("clust_plotting_obj('"+self.in_dir+"')")

#####################
class clust_obj(object):
	def __init__(self, in_dir):
		self.in_dir = in_dir
		self.load_labels()
		self.load_expression_stats()
		self.load_marker_table()
		self.load_clust_genes()
		self.load_enrich_genes()
		self.create_clust_plot()

	def load_labels(self):
		label_table = load_relative_table(self.in_dir, "sample_clustering_and_summary/sample_k_means_groups.tsv")
		self.samples = label_table[:,0].tolist()
		self.sample_idx_dict = {key:value for value, key in enumerate(self.samples)}
		self.labels = np.array(label_table[:,1],dtype=float)
		self.cluster_lists = get_sample_k_lists(self.labels.tolist())

	def load_marker_table(self):
		print("loading the cluster level marker gene annotations")
		self.best_markers = load_relative_table(self.in_dir, "sample_clustering_and_summary/significance/high_markers/best_sorted_markers.tsv", out_type = "pandas", no_head = True)
		self.all_markers = load_relative_table(self.in_dir, "sample_clustering_and_summary/significance/high_markers/marker_gene_annotations.tsv", out_type = "pandas")

	def load_expression_stats(self):
		print("loading the global & cluster expression metrics")
		self.global_non_zero_mean = load_relative_table(self.in_dir, "sample_clustering_and_summary/expression_metrics/global_non_zero_mean.tsv", out_type = "pandas")
		self.global_percent_express = load_relative_table(self.in_dir, "sample_clustering_and_summary/expression_metrics/global_percent_express.tsv", out_type = "pandas")
		self.cluster_percent_express = load_relative_table(self.in_dir, "sample_clustering_and_summary/expression_metrics/k_group_percent_express.tsv", out_type = "pandas")
		self.cluster_non_zero_mean_expression = load_relative_table(self.in_dir, "sample_clustering_and_summary/expression_metrics/non_zero_mean_expression.tsv", out_type = "pandas")
		self.ratio_non_zero_mean_expression = load_relative_table(self.in_dir, "sample_clustering_and_summary/expression_metrics/ratio_non_zero_mean_expression.tsv", out_type = "pandas")
		self.ratio_percent_express = load_relative_table(self.in_dir, "sample_clustering_and_summary/expression_metrics/ratio_percent_express.tsv", out_type = "pandas")
		self.cluster_express_mean = load_relative_table(self.in_dir, "sample_clustering_and_summary/k_group_means.tsv", out_type = "pandas")
		self.cluster_express_sd = load_relative_table(self.in_dir, "sample_clustering_and_summary/k_group_sd.tsv", out_type = "pandas")

	def load_clust_genes(self):
		self.clust_genes = load_relative_table(self.in_dir, "sample_clustering_and_summary/genes_used_for_clustering.txt", out_type = "numpy")

	def load_enrich_genes(self):
		print("loading the genes that are enriched in each group & their pathway analyses")
		self.enrich_table = load_relative_table(self.in_dir, "sample_clustering_and_summary/sample_var_enrichment_Zscores.txt", out_type = "pandas")
		self.enrich_bool_table = load_relative_table(self.in_dir, "sample_clustering_and_summary/significance/significant_and_enriched_boolean_table.tsv", out_type = "pandas", no_head = True)
		self.significance_table = load_relative_table(self.in_dir, "sample_clustering_and_summary/significance/groups_1way_anova_results.tsv", out_type = "pandas")
		self.pathway_importance = load_relative_table(self.in_dir, "sample_clustering_and_summary/significance/individual_class_importance.tsv", out_type = "pandas")
		self.pathway_neg_log_p_vals = load_relative_table(self.in_dir, "sample_clustering_and_summary/significance/combined_neg_log10p_gprofiler.tsv", out_type = "pandas")

	def create_clust_plot(self):
		self.plt_info = clust_plotting_obj(self.in_dir)

	def __str__(self):
		out_str = "PyMINEr clustering analyses"
		out_str += "\nObjects of interest:"
		out_str += "\n\nsample level info:"
		out_str += "\n\t<object>.samples: the vector of sample IDs"
		out_str += "\n\t<object>.sample_idx_dict: a dictionary with the sample names as keys & their index in the \n\t\tdataset as the value"
		out_str += "\n\t<object>.labels: the cluster labels in the form of cluster number"
		out_str += "\n\t<object>.cluster_lists: a list of lists where the first level is the cluster, & the \n\t\tsecond level is the sample indexes that belong to that cluster"
		out_str += "\n\nmarker gene info:"
		out_str += "\n\t<object>.best_markers: The top n marker genes that define each cluster."
		out_str += "\n\t<object>.all_markers: A table of all genes, how good of markers they make, & what their \n\t\tbest cluster mapping is."
		out_str += "\n\nexpression statistics info:"
		out_str += "\n\t<object>.global_non_zero_mean: in the whole dataset what is the non-zero mean?"
		out_str += "\n\t<object>.cluster_non_zero_mean_expression: within each cluster, what is the non-zero mean?"
		out_str += "\n\t<object>.ratio_non_zero_mean_expression: for each cluster, what's the ratio beteween its \n\t\tnon-zero mean and the global non-zero mean?"
		out_str += "\n\t<object>.global_percent_express: in the whole dataset, what's the percentage of \n\t\tnon-zero expression?"
		out_str += "\n\t<object>.cluster_percent_express: within each cluster, what's the percentage of \n\t\tnon-zero expression?"
		out_str += "\n\t<object>.ratio_percent_express: for each cluster, what's the ratio of the \n\t\tnon-zero percentage & the global non-zero percent expressed?"
		out_str += "\n\t<object>.cluster_express_mean: the mean expression of genes in each cluster"
		out_str += "\n\t<object>.cluster_express_sd: the standard deviation of gene expression in each cluster"
		out_str += "\n\ncluster expression enrichment info:"
		out_str += "\n\t<object>.enrich_table: the cluster level Z-score table for expression"
		out_str += "\n\t<object>.enrich_bool_table: a boolean T/F table for all genes and clusters for \n\t\twhether or not a gene was significantly enriched in that cluster"
		out_str += "\n\t<object>.significance_table: the anova significance table. (global F-statistic, \n\t\tnot pairwise comparisons)"
		out_str += "\n\t<object>.pathway_importance: A table of pathway importance factors (rows) for each \n\t\tcluster (columns). Sort a column to see what pathways are most informative/unique\n\t\tin that specific cluster."
		out_str += "\n\t<object>.pathway_neg_log_p_vals: A table of the -log10(p-values) for pathways (rows) \n\t\tfor each cluster (column), sorted by PyMINEr's information ranking system \n\t\t(based on KL-divergence)"
		out_str += "\n\nclustering and plotting info:"
		out_str += "\n\t<object>.plt_info: An object that contains detailed plotting info. Print that \n\t\tobject for it's own help guide."
		# out_str += "\n\t<object>.samples:"
		# out_str += "\n\t<object>.samples:"
		# out_str += "\n\t<object>.samples:"
		return(out_str)

	def __repr__(self):
		return("clust_obj('"+self.in_dir+"')")
##################

class gene_annotations_obj(object):
	def __init__(self, in_dir):
		self.in_dir = in_dir
		self.load_anno()

	def load_anno(self):
		self.annotations = load_relative_table(self.in_dir, "annotations.tsv", out_type = "pandas")
		ortho_file = os.path.join(self.in_dir,"human_orthologues.tsv")
		if os.path.isfile(ortho_file):
			self.human_ortho = load_relative_table(self.in_dir, "human_orthologues.tsv", out_type = "pandas")
		else:
			self.human_ortho = None

	def __str__(self):
		out_str = "PyMINEr gene annotations"
		out_str += "\nObjects of interest:"
		out_str += "\n\t<object>.annotations: A table of gene level annotations, including Ensembl gene \n\t\tmapping & gene descriptions."
		out_str += "\n\t<object>.human_ortho: If you didn't run a human sample, there should be a table \n\t\tannotating the human orthologues. Otherwise, this is None."
		return(out_str)

	def __repr__(self):
		return("gene_annotations_obj('"+self.in_dir+"')")

##################

class pyminer_obj(object):
	def __init__(self, in_dir):
		self.in_dir = in_dir
		self.gene_anno = gene_annotations_obj(in_dir)
		self.clust = clust_obj(in_dir)
		self.goi = goi_obj(in_dir)
		self.network = network_obj(in_dir)
		self.ap = ap_obj(in_dir)
		
	def set_dir(self, new_in_dir):
		self.in_dir = new_in_dir
		self.gene_anno.in_dir = new_in_dir
		self.clust.in_dir = new_in_dir
		self.goi.in_dir = new_in_dir
		self.network.in_dir = new_in_dir
		self.ap.in_dir = new_in_dir

	def __str__(self):
		out_str = "PyMINEr object:\n"
		out_str += "\nEach of the elements contained within the PyMINEr object is it's own class\n\t\tYou can print each of these objects to get the help section on it."
		out_str += "\n\t<object>.gene_anno: an object that contains tables annotating the input genes"
		out_str += "\n\t<object>.clust: This object has lots of info on the clusters, differential \n\t\texpression, pathway analysis, etc. just do print(<object>.clust) to get the details!"
		out_str += "\n\t<object>.goi: If you used genes of interest, the direcotry of the output \n\t\tplots and the shortest-paths of all other genes to them are located here."
		out_str += "\n\t<object>.network: The co-expression network and details on gene-module \n\t\tusage, module pathway analysis, etc"
		out_str += "\n\t<object>.ap: Details of the autocrine/paracrine signaling between clusters \n\t\tand pathway analysis of those signaling networks."
		#out_str += "\n\t<object>.: "
		out_str += "\n\nThere is also the PyMINEr website file that provides a walk-through of the results:\n"+os.path.join(self.in_dir,"PyMINEr_summary.html")+"\n\n"
		#out_str += "\n:"+
		return(out_str)

	def __repr__(self):
		return("pyminer_obj("+self.in_dir+")")



#!/usr/bin/env python3

##import dependency libraries
import sys,time,glob,os,pickle,fileinput,argparse, random
from time import sleep
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
import argparse
try:
    from pyminer.common_functions import *
except:
    from common_functions import *
##############################################################

##############################################################
##############################################################



def get_name_from_path(in_path):
    temp_name = in_path.split('/')
    temp_name = temp_name[-1]
    temp_name = temp_name[:-4]
    return(temp_name)

def get_gene_lists_from_files(in_files):
    all_gene_list_files = in_files.split(',')
    cell_type_ids = []
    for gene_list in all_gene_list_files:
        cell_type_ids.append(get_name_from_path(gene_list))

    ## get the gene lists for each cell type
    gene_lists = []
    gene_ids = []
    for cell in all_gene_list_files:
        temp_genes = read_file(cell)
        gene_lists.append(temp_genes)
        gene_ids += temp_genes
    gene_ids = list(set(gene_ids))
    return(cell_type_ids, gene_ids, gene_lists)

def get_gene_lists_from_bool_table(in_file):
    """
    here, we take in the boolean table with columns corresponding
    to the cell type names, with the body of the table consisting of
    True or False. This will be directly translated into a boolean numpy array
    so syntax is important in this array.
    """
    bool_table = read_table(in_file)
    cell_type_ids = bool_table[0][1:]
    bool_table = np.array(bool_table)
    gene_ids = list(set(bool_table[1:,0].tolist()))
    gene_lists = []
    ## set up the empty lists
    for i in range(1, np.shape(bool_table)[1]):
        gene_lists.append([])
    ## populate the 
    for i in range(1,np.shape(bool_table)[0]):
        ## go through each row in the table (barring the )
        for j in range(1, np.shape(bool_table)[1]):
            if bool_table[i,j]=='True':
                gene_lists[j-1].append(bool_table[i,0])
    
    ## make them all upper case
    for i in range(0,len(gene_ids)):
        gene_ids[i]=str(gene_ids[i]).upper()

    for i in range(0,len(gene_lists)):
        for j in range(0,len(gene_lists[i])):
            gene_lists[i][j]=str(gene_lists[i][j]).upper()

    return(cell_type_ids, gene_ids, gene_lists)


def convert_to_ensp(organism, gp, in_genes, gene_lists):
    temp_lookup={}
    gene_symbol_dict = {}
    for g in in_genes:
        temp_lookup[g] = []
    results = gp.gconvert(in_genes, 
        organism = organism, target='ENSP')
    for r in range(1,len(results)):
        ## the original ID
        g = results[r][1]
        ## the current orthologue list
        try:
            temp_lookup[g]
        except:
            pass
            # print('weird mapping event:')
            # print(results[r])
        else:
            if results[r][4] != None:
                gene_symbol_dict[results[r][3]]=results[r][4]
            if results[r][3] != None:
                temp_list = temp_lookup[g]
                temp_list += [results[r][3]]
                temp_lookup[g]=temp_list

    gene_lists = convert_list_of_lists_by_dict(gene_lists,temp_lookup)
    new_gene_ids = []
    for k in list(temp_lookup.keys()):
        new_gene_ids += temp_lookup[k]
    new_gene_ids= list(set(new_gene_ids))

    for gene in new_gene_ids:
        if gene not in gene_symbol_dict:
            gene_symbol_dict[gene]="None"


    return(new_gene_ids, gene_lists, gene_symbol_dict)


def get_orthologue_dicts(gp, in_genes,source_organism, target_organism, orthologues = True, ensp_convert = False):
    temp_org_to_human = {}

    ## each gene will get a list for its orthologues
    for g in in_genes:
        temp_org_to_human[g] = []

    results = gp.gorth(in_genes, 
        source_organism = source_organism, 
        target_organism=target_organism,
        numeric_ns=True)


    for r in range(1,len(results)):

        ## the original ID
        g = results[r][1]
        ## the current orthologue list
        try:
            temp_org_to_human[g]
        except:
            pass
            # print('weird mapping event:')
            # print(results[r])
        else:
            if results[r][4] != None:
                temp_list = temp_org_to_human[g]
                temp_list += [results[r][4]]
                temp_org_to_human[g]=temp_list

    human_to_temp_org = {}
    return(temp_org_to_human,human_to_temp_org)

def convert_list_of_lists_by_dict(gene_lists,temp_org_to_human):
    ## set up the new gene list
    new_gene_lists = []
    for i in range(len(gene_lists)):
        new_gene_lists.append([])
    ## convert the original gene lists to the IDs
    for i in range(len(gene_lists)):
        for j in range(len(gene_lists[i])):
            new_gene_lists[i]+=temp_org_to_human[gene_lists[i][j]]
    return(new_gene_lists)


def update_converted_genes(temp_org_to_human, gene_ids, gene_lists):
    new_gene_ids = []
    for k in list(temp_org_to_human.keys()):
        new_gene_ids += temp_org_to_human[k]
    new_gene_ids= list(set(new_gene_ids))
    if None in new_gene_ids:
        new_gene_ids.pop(None)

    new_gene_lists=convert_list_of_lists_by_dict(gene_lists,temp_org_to_human)

    return(new_gene_ids, new_gene_lists)

def get_receptors_and_secreted(organism, gp, gene_ids):
    secreted_go = ['GO:0005615','GO:0005576','GO:0044421']
    receptor_go = ["GO:0009897","GO:0031232","GO:0031233","GO:0071575"," GO:0098591","GO:0031362","GO:0098567","GO:0009986","GO:0005886","GO:0042923","GO:0016021"]#,'GO:0005904']

    results = gp.gconvert(gene_ids, 
        organism = organism, target='GO')


    # print(results[0])
    # print(results[1])
    # print(results[2])
    # print(results[3])

    all_go_terms = []
    receptor_list = []
    secreted_list = []
    for i in range(1,len(results)):
        if results[i][3] not in all_go_terms:
            all_go_terms.append(results[i][3])
        if results[i][3] in receptor_go:
            receptor_list.append(results[i][1])
        if results[i][3] in secreted_go:
            secreted_list.append(results[i][1])
    receptor_list = list(set(receptor_list))
    secreted_list = list(set(secreted_list))
    return(receptor_list, secreted_list)

def filter_enriched_lists_for_subcellular_localization(receptor_list, secreted_list, gene_lists):
    receptor_gene_lists = []
    secreted_gene_lists = []
    for g in range(len(gene_lists)):
        receptor_gene_lists.append([])
        secreted_gene_lists.append([])

    for i in range(0,len(gene_lists)):
        for j in range(0,len(gene_lists[i])):
            if gene_lists[i][j] in receptor_list:
                receptor_gene_lists[i].append(gene_lists[i][j])
            if gene_lists[i][j] in secreted_list:
                secreted_gene_lists[i].append(gene_lists[i][j])
    return(receptor_gene_lists,secreted_gene_lists)

def clean_names(in_vect):
    name1 = in_vect[0].split('.')
    name2 = in_vect[1].split('.')
    name1 = name1[-1]
    name2 = name2[-1]
    in_vect[0] = name1
    in_vect[1] = name2
    return(in_vect)

def subset_action_table(receptor_or_secreted_list,db_file):
    print('this may take a minute....')
    ## construct a dictionary for the receptor or secreted list
    receptor_or_secreted_dict = {j:i for i, j in enumerate(receptor_or_secreted_list)}
    print(len(list(receptor_or_secreted_dict.keys())), 'receptor_or_secreted_dict')
    all_ids = []
    line_count = 0
    out_table = []
    activation_dict = {}
    inhibition_dict = {}
    other_dict = {}
    first = True
    for line in fileinput.input(db_file):
        line_count+=1
        if line_count % 1e5 == 0:
            print('\t',line_count,'\t',len(out_table))
        temp_line = strip_split(line)
        if first:
            first = False
            #out_table.append(temp_line) 
        else:
            temp_line = clean_names(temp_line)


            ## check if the two genes are in the receptor or secreted list
            # holy shit this way of checking if something is in a dictionary is so much
            # faster than checking if they're in a list... good to know...
            # I'm talking like, orders of magnitude faster...
            # I've wasted so much time waiting on this step. Now it takes seconds o...
            cont1 = True
            try:
                receptor_or_secreted_dict[temp_line[0]]
            except:
                cont1=False
            if cont1:
                all_ids.append(temp_line[0])

            cont2 = True
            try:
                receptor_or_secreted_dict[temp_line[1]]
            except:
                cont2=False
            if cont2:
                all_ids.append(temp_line[1])

            if cont1 and cont2:
            #if temp_line[0] in receptor_or_secreted_list and temp_line[1] in receptor_or_secreted_list:
                temp_id = temp_line[0]+":"+temp_line[1]
                if temp_line[2] == 'binding': 
                    out_table.append(temp_line)
                elif temp_line[2] == 'activation':
                    activation_dict[temp_id] = True
                elif temp_line[2] == 'inhibition':
                    inhibition_dict[temp_id] = True
                else:
                    if temp_id not in other_dict:
                        other_dict[temp_id]=temp_line[2]
                    else:
                        if other_dict[temp_id] != 'None':
                            temp_thingy = other_dict[temp_id]
                            temp_thingy = ','+temp_line[2]
                        if temp_thingy[0]==',':
                            temp_thingy=temp_thingy[1:]
                        other_dict[temp_id] = temp_thingy

    fileinput.close()
    all_ids = list(set(all_ids))
    return(out_table, activation_dict, inhibition_dict, other_dict,all_ids)

def annotate_interactions_for_activation_and_inhibition(action_table, activation_dict, inhibition_dict, other_dict):
    print('annotating the interactions for directionality and type of action....')
    action_table[0] = action_table[0][:11]+['activation/inhibition','other_action']+action_table[0][11:]
    # print(action_table[0])
    for i in range(1,len(action_table)):
    #     for l in enumerate(action_table[i]):
    #         print(l)
        temp_id = str(action_table[i][9])+":"+str(action_table[i][10])
        activation_term = 'None'
        other_action_term = 'None'
        if temp_id in activation_dict:
            activation_term = 'activation'
        if temp_id in inhibition_dict:
            activation_term = 'inhibition'
        if temp_id in other_dict:
            other_action_term = other_dict[temp_id]
        action_table[i] = action_table[i][:11]+[activation_term,other_action_term]+action_table[i][11:]
    return(action_table)



def gene_lists_to_symbol_lists(gene_lists, gene_symbol_dict):
    gene_symbol_lists = []
    for c in range(len(gene_lists)):
        gene_symbol_lists.append([])
        for gene in gene_lists[c]:
            temp_symbol = gene_symbol_dict[gene]
            if temp_symbol not in gene_symbol_lists[c]:
                gene_symbol_lists[c].append(temp_symbol)
    return(gene_symbol_lists)

def gene_symbol_lists_to_dicts(gene_symbol_lists):
    gene_symbol_list_dicts = []
    for i in range(0,len(gene_symbol_lists)):
        gene_symbol_list_dicts.append({j:w for w, j in enumerate(gene_symbol_lists[i])})
    return(gene_symbol_list_dicts)



##########################################################

def get_empty_interaction_lists(gene_lists):
    interaction_lists = []
    for i in range(0,len(gene_lists)):
        interaction_lists.append([])
        for j in range(0,len(gene_lists)):
            interaction_lists[i].append([])
    return(interaction_lists)

    
def get_cell_types_gene_enriched_in(gene_list_dict, gene_lists, cell_type_ids, gene):
    gene_cell_types = []
    for i in range(0,len(cell_type_ids)):
        cur_cell_type = True
        try:
            gene_list_dict[i][gene]
        except:
            cur_cell_type = False
        if cur_cell_type:
            gene_cell_types.append(cell_type_ids[i])
    # ## just for trouble shooting
    # if gene =="ENSP00000357461" or gene == "ENSP00000383938":
    #     print('\n\n',gene_cell_types,'\n\n')
    return(gene_cell_types)


##########################################################
## get gprofiler results from each of the interaction lists
def key_in_dict(key, in_dict):
    in_it = True
    try:
        in_dict[key]
    except:
        in_it = False
    return(in_it)


def get_gprofiler_on_interaction_lists(organism, gp, cell_type_ids, interaction_lists, bg = None):
    print('getting the pathway analysis for each cell type interactions...')
    output_gprofiler = [['cell_type_1','cell_type_2']]
    for i in range(0,len(interaction_lists)):
        ## wait for a random amount of time so that we don't overload the server
        random_time_lag = np.random.randint(low=25,high=100)/100
        sleep(random_time_lag)
        for j in range(i,len(interaction_lists)):
            print('\t',cell_type_ids[i],cell_type_ids[j],":",len(interaction_lists[i][j]),'genes')
            if bg == None:
                temp_gprofile_results = gp.gprofile(interaction_lists[i][j], organism = organism)
            else:
                temp_gprofile_results = gp.gprofile(interaction_lists[i][j], custom_bg = bg, organism = organism)
            if i==0 and j == 0 and temp_gprofile_results!=[]:
                output_gprofiler[0] += temp_gprofile_results[0]
            for q in range(1,len(temp_gprofile_results)):
                output_gprofiler.append([cell_type_ids[i],cell_type_ids[j]]+temp_gprofile_results[q])
    return(output_gprofiler)

# def see_if_symbols_interact(symbol_interaction_dict, symbol1,symbol2):
#     temp_key = str(symbol1)+":"+str(symbol2)
#     if key_in_dict(temp_key,symbol_interaction_dict):
#         return(True)
#     temp_key = str(symbol2)+":"+str(symbol1)
#     if key_in_dict(temp_key,symbol_interaction_dict):
#         return(True)
#     return(False)


def filter_and_annoatate_genes(symbol_interaction_dict, 
                               gene_symbol_lists, 
                               cell_type_idx_dict, 
                               gene_symbol_list_dicts, 
                               in_gene_list, 
                               cell_type_1, 
                               cell_type_2, 
                               test_time=False):
    if cell_type_1 != cell_type_2:
        paracrine = True
    else:
        paracrine = False
    in_gene_list = in_gene_list.split(',')
    if len(in_gene_list)==1:
        return("None")
    cell_type_1_idx = cell_type_idx_dict[cell_type_1]
    cell_type_2_idx = cell_type_idx_dict[cell_type_2]
    #cell_type_enriched_ids_1=gene_symbol_lists[cell_type_1_idx]
    #cell_type_enriched_ids_2=gene_symbol_lists[cell_type_2_idx]
    cell_type_enriched_dict_1=gene_symbol_list_dicts[cell_type_1_idx]
    cell_type_enriched_dict_2=gene_symbol_list_dicts[cell_type_2_idx]
    
    ## first make all fully crossed interaction_pairs
    if test_time:
        print('making pairs')
        start = time()
    fully_crossed_possible_pairs = []
    for i in range(0,len(in_gene_list)):
        for j in range(i,len(in_gene_list)):
            #if in_gene_list[j]+":"+in_gene_list[i] not in fully_crossed_possible_pairs:
            fully_crossed_possible_pairs.append(in_gene_list[i]+":"+in_gene_list[j])
            #fully_crossed_possible_pairs.append(in_gene_list[j]+":"+in_gene_list[i])
    fully_crossed_possible_pairs = list(set(fully_crossed_possible_pairs))
    filtered_pairs = []
    if test_time:
        print('\t',time()-start)
        print('\tlooking at fully_crossed_possible_pairs in interaction dict')
        start = time()
    for i in range(0,len(fully_crossed_possible_pairs)):
        cont = True
        try:
            symbol_interaction_dict[fully_crossed_possible_pairs[i]]
        except:
            cont = False
        if cont:#fully_crossed_possible_pairs[i] in symbol_interaction_dict:
            filtered_pairs.append(fully_crossed_possible_pairs[i])
    #print(filtered_pairs)
    if test_time:
        print('\t\t',time()-start)
        print('\tannotating source')
        start = time()
    final_output_list = []
    for i in range(0,len(filtered_pairs)):
        ## format of this will be cell_type_1:gene1;cell_type2:gene2,... for each interaction contained within the genes in this pathway
        temp_interaction = filtered_pairs[i]
        temp_interaction = temp_interaction.split(':')
        gene_source_1 = ''
        gene_source_2 = ''
        ## figure out which cell type is producing each within the pair 
        if key_in_dict(temp_interaction[0], cell_type_enriched_dict_1):
            #print(temp_interaction[0],"found in",cell_type_1,cell_type_enriched_ids_1)
            gene_source_1+=cell_type_1
        if key_in_dict(temp_interaction[0], cell_type_enriched_dict_2):
            #temp_interaction[0] in cell_type_enriched_ids_2:
            #print(temp_interaction[0],"found in",cell_type_2,cell_type_enriched_ids_2)
            if gene_source_1!='':
                gene_source_1+='&'
            gene_source_1+=cell_type_2
        #print(gene_source_1)
        if key_in_dict(temp_interaction[1], cell_type_enriched_dict_1):
        #if temp_interaction[1] in cell_type_enriched_ids_1:
            #print(temp_interaction[1],"found in",cell_type_1,cell_type_enriched_ids_1)
            gene_source_2+=cell_type_1
        if key_in_dict(temp_interaction[1], cell_type_enriched_dict_2):
        #if temp_interaction[1] in cell_type_enriched_ids_2:
            #print(temp_interaction[1],"found in",cell_type_2,cell_type_enriched_ids_2)
            if gene_source_2!='':
                gene_source_2+='&'
            gene_source_2+=cell_type_2
        ##print(gene_source_2)
        ## double check that there was a gene source found
        if gene_source_1 == '':
            print("couldn't find gene_source_1")
        if gene_source_2 == '':
            print("couldn't find gene_source_2")
        ## format the outputs for this line
        if paracrine:
            temp_full_interaction_description = gene_source_1+":"+temp_interaction[0]+";"+gene_source_2+":"+temp_interaction[1]
            #print(temp_full_interaction_description)
        else:
            temp_full_interaction_description = cell_type_1+":"+temp_interaction[0]+";"+cell_type_2+":"+temp_interaction[1]
        if paracrine and gene_source_1!=gene_source_2:
            final_output_list.append(temp_full_interaction_description)
        elif not paracrine:
            final_output_list.append(temp_full_interaction_description)
    if test_time:
        print('\t\t',time()-start)
    ## and return the final line
    if final_output_list == []:
        return('None')
    return(','.join(final_output_list))



def annotate_gprofiler_output(symbol_interaction_dict, 
                              gene_symbol_lists, 
                              cell_type_idx_dict, 
                              gene_symbol_list_dicts,
                              output_gprofiler):
    print('\nannotating the gprofiler results\nThis might take a minute...')
    new_output_gprofiler = [output_gprofiler[0]+[output_gprofiler[0][-1]]]
    new_output_gprofiler[0][-2] = 'filtered_annotated_gene_list'
    first = True
    count=0
    total_length = len(output_gprofiler)
    for line in output_gprofiler:
        count+=1
        if count%500 == 0:
            print('\tline:',count,'\tpercent:',100*count/total_length,'%')
        if first:
            first = False
            pass
        else:
            ## put the cell types and annotations into the line, then append it to the new output
            ## but only add it for either all autocrine interactions, or for paracrine interactions
            ## only put it into the output if there are 
            temp_line = line[:]
            temp_line.append(temp_line[-1])
            temp_line[-2] = filter_and_annoatate_genes(symbol_interaction_dict, 
                                                       gene_symbol_lists, 
                                                       cell_type_idx_dict, 
                                                       gene_symbol_list_dicts,
                                                       temp_line[-1],
                                                       temp_line[0],
                                                       temp_line[1])
            if temp_line[-2] != "None":
                new_output_gprofiler.append(temp_line)
    return(new_output_gprofiler)

    
def get_subset(out_table, interaction_title_line, out_file,term_a,term_b):
    temp_file_str = out_file
    rm(temp_file_str)
    extra_cell_file = open(temp_file_str,'a')
    temp_title_line='\t'.join(interaction_title_line[0])
    extra_cell_file.write(temp_title_line+'\n')
    #extracellular_extracellular_interactions = interaction_title_line[:]
    for i in range(1,len(out_table)):
        if (out_table[i][4]  == term_a and out_table[i][8] == term_b) or (out_table[i][4]  == term_b and out_table[i][8] == term_a):
            temp_line = out_table[i][:]
            temp_line = '\t'.join(temp_line)
            if i!=len(out_table)-1:
                temp_line+='\n'
            extra_cell_file.write(temp_line)
    extra_cell_file.close()
    return




def get_unique_ids(out_line):
    gene_1 = out_line[2]
    cell_type_1 = out_line[1]
    gene_2 = out_line[6]
    cell_type_2 = out_line[5]
    inuqe_id_1 = cell_type_1+"|"+gene_1+":"+cell_type_2+"|"+gene_2
    inuqe_id_2 = cell_type_2+"|"+gene_2+":"+cell_type_1+"|"+gene_1
    return(inuqe_id_1,inuqe_id_2)

##############################################################
##############################################################

def pyminer_cell_signals_main(gene_lists,
                              gene_table,
                              out_dir,
                              organism = "hsapiens",
                              species_codes = False,
                              convert_to_human = False,
                              stringdb_dir = "/usr/local/lib/cell_signals/"):
    organism_action_files = {
        'hsapiens':'9606.protein.actions.v11.0.txt',
        'mmusculus':'10090.protein.actions.v11.0.txt'
    }


    if organism not in organism_action_files:
        convert_to_human = True

    ## check that we have gprofiler installed
    try:
        from gprofiler import GProfiler
    except:
        sys.exit('please install gprofiler; try: pip3 install gprofiler-official')
    else:
        from gprofiler import GProfiler

    gp = GProfiler('PyMINEr_'+str(random.randint(0,int(1e6))), want_header = True)

    ##############################################################
    ##############################################################
    if species_codes:
        print("""
    Ensembl

        aaegypti — Aedes aegypti
        acarolinensis — Anolis carolinensis
        agambiae — Anopheles gambiae
        amelanoleuca — Ailuropoda melanoleuca
        amexicanus — Astyanax mexicanus
        aplatyrhynchos — Anas platyrhynchos
        btaurus — Bos taurus
        celegans — Caenorhabditis elegans
        cfamiliaris — Canis familiaris
        choffmanni — Choloepus hoffmanni
        cintestinalis — Ciona intestinalis
        cjacchus — Callithrix jacchus
        cporcellus — Cavia porcellus
        csabaeus — Chlorocebus sabaeus
        csavignyi — Ciona savignyi
        dmelanogaster — Drosophila melanogaster
        dnovemcinctus — Dasypus novemcinctus
        dordii — Dipodomys ordii
        drerio — Danio rerio
        ecaballus — Equus caballus
        eeuropaeus — Erinaceus europaeus
        etelfairi — Echinops telfairi
        falbicollis — Ficedula albicollis
        fcatus — Felis catus
        gaculeatus — Gasterosteus aculeatus
        ggallus — Gallus gallus
        ggorilla — Gorilla gorilla
        gmorhua — Gadus morhua
        hsapiens — Homo sapiens
        itridecemlineatus — Ictidomys tridecemlineatus
        lafricana — Loxodonta africana
        lchalumnae — Latimeria chalumnae
        loculatus — Lepisosteus oculatus
        mdomestica — Monodelphis domestica
        meugenii — Macropus eugenii
        mfuro — Mustela putorius furo
        mgallopavo — Meleagris gallopavo
        mlucifugus — Myotis lucifugus
        mmulatta — Macaca mulatta
        mmurinus — Microcebus murinus
        mmusculus — Mus musculus
        nleucogenys — Nomascus leucogenys
        oanatinus — Ornithorhynchus anatinus
        oaries — Ovis aries
        ocuniculus — Oryctolagus cuniculus
        ogarnettii — Otolemur garnettii
        olatipes — Oryzias latipes
        oniloticus — Oreochromis niloticus
        oprinceps — Ochotona princeps
        pabelii — Pongo abelii
        panubis — Papio anubis
        pcapensis — Procavia capensis
        pformosa — Poecilia formosa
        pmarinus — Petromyzon marinus
        psinensis — Pelodiscus sinensis
        ptroglodytes — Pan troglodytes
        pvampyrus — Pteropus vampyrus
        rnorvegicus — Rattus norvegicus
        saraneus — Sorex araneus
        scerevisiae — Saccharomyces cerevisiae
        sharrisii — Sarcophilus harrisii
        sscrofa — Sus scrofa
        tbelangeri — Tupaia belangeri
        tguttata — Taeniopygia guttata
        tnigroviridis — Tetraodon nigroviridis
        trubripes — Takifugu rubripes
        tsyrichta — Tarsius syrichta
        ttruncatus — Tursiops truncatus
        vpacos — Vicugna pacos
        xmaculatus — Xiphophorus maculatus
        xtropicalis — Xenopus tropicalis

    Ensembl Genomes Fungi

        aclavatus — Aspergillus clavatus
        aflavus — Aspergillus flavus
        afumigatus — Aspergillus fumigatus
        afumigatusa1163 — Aspergillus fumigatusa1163
        agossypii — Ashbya gossypii
        anidulans — Aspergillus nidulans
        aniger — Aspergillus niger
        aoryzae — Aspergillus oryzae
        aterreus — Aspergillus terreus
        bcinerea — Botrytis cinerea
        bgraminis — Blumeria graminis f. sp. hordei DH14
        cgloeosporioides — Colletotrichum gloeosporioides
        cgraminicola — Colletotrichum graminicola
        chigginsianum — Colletotrichum higginsianum
        cneoformans — Cryptococcus neoformans
        corbiculare — Colletotrichum orbiculare
        dseptosporum — Dothistroma septosporum
        fculmorum — Fusarium culmorum UK99
        ffujikuroi — Fusarium fujikuroi
        fgraminearum — Fusarium graminearum
        foxysporum — Fusarium oxysporum
        fpseudograminearum — Fusarium pseudograminearum
        fsolani — Fusarium solani
        fverticillioides — Fusarium verticillioides
        ggraminis — Gaeumannomyces graminis
        kpastoris — Komagataella pastoris
        lmaculans — Leptosphaeria maculans
        mlaricipopulina — Melampsora larici-populina 98AG31
        moryzae — Magnaporthe oryzae
        mpoae — Magnaporthe poae
        mviolaceum — Microbotryum violaceum p1A1 Lamole
        ncrassa — Neurospora crassa
        nfischeri — Neosartorya fischeri
        pgraminis — Puccinia graminis
        pgraminisug99 — Puccinia graminis Ug99
        pnodorum — Phaeosphaeria nodorum
        pstriiformis — Puccinia striiformis f. sp. tritici PST-130 str. Race 130
        pteres — Pyrenophora teres f. teres 0-1
        ptriticina — Puccinia triticina
        ptriticirepentis — Pyrenophora tritici-repentis Pt-1C-BFP
        scerevisiae — Saccharomyces cerevisiae
        scryophilus — Schizosaccharomyces cryophilus
        sjaponicus — Schizosaccharomyces japonicus
        soctosporus — Schizosaccharomyces octosporus
        spombe — Schizosaccharomyces pombe
        sreilianum — Sporisorium reilianum SRZ2
        ssclerotiorum — Sclerotinia sclerotiorum
        tmelanosporum — Tuber melanosporum
        treesei — Trichoderma reesei
        tvirens — Trichoderma virens
        umaydis — Ustilago maydis
        vdahliae — Verticillium dahliae
        vdahliaejr2 — Verticillium dahliae JR2
        ylipolytica — Yarrowia lipolytica
        ztritici — Zymoseptoria tritici

    Ensembl Genomes Metazoa

        aaegypti — Aedes aegypti
        acephalotes — Atta cephalotes
        adarlingi — Anopheles darlingi
        agambiae — Anopheles gambiae
        aglabripennis — Anoplophora glabripennis
        amellifera — Apis mellifera
        apisum — Acyrthosiphon pisum
        aqueenslandica — Amphimedon queenslandica
        avaga — Adineta vaga
        bantarctica — Belgica antarctica
        bimpatiens — Bombus impatiens
        bmalayi — Brugia malayi
        bmori — Bombyx mori
        cbrenneri — Caenorhabditis brenneri
        cbriggsae — Caenorhabditis briggsae
        celegans — Caenorhabditis elegans
        cgigas — Crassostrea gigas
        cjaponica — Caenorhabditis japonica
        cquinquefasciatus — Culex quinquefasciatus
        cremanei — Caenorhabditis remanei
        cteleta — Capitella teleta
        dananassae — Drosophila ananassae
        derecta — Drosophila erecta
        dgrimshawi — Drosophila grimshawi
        dmelanogaster — Drosophila melanogaster
        dmojavensis — Drosophila mojavensis
        dpersimilis — Drosophila persimilis
        dplexippus — Danaus plexippus
        dponderosae — Dendroctonus ponderosae
        dpseudoobscura — Drosophila pseudoobscura
        dpulex — Daphnia pulex
        dsechellia — Drosophila sechellia
        dsimulans — Drosophila simulans
        dvirilis — Drosophila virilis
        dwillistoni — Drosophila willistoni
        dyakuba — Drosophila yakuba
        hmelpomene — Heliconius melpomene
        hrobusta — Helobdella robusta
        iscapularis — Ixodes scapularis
        lanatina — Lingula anatina
        lcuprina — Lucilia cuprina
        lgigantea — Lottia gigantea
        lloa — Loa loa
        lsalmonis — Lepeophtheirus salmonis
        mcinxia — Melitaea cinxia
        mdestructor — Mayetiola destructor
        mleidyi — Mnemiopsis leidyi
        mscalaris — Megaselia scalaris
        nvectensis — Nematostella vectensis
        nvitripennis — Nasonia vitripennis
        obimaculoides — Octopus bimaculoides
        ovolvulus — Onchocerca volvulus
        phumanus — Pediculus humanus
        ppacificus — Pristionchus pacificus
        rprolixus — Rhodnius prolixus
        sinvicta — Solenopsis invicta
        smansoni — Schistosoma mansoni
        smaritima — Strigamia maritima
        smimosarum — Stegodyphus mimosarum
        spurpuratus — Strongylocentrotus purpuratus
        sratti — Strongyloides ratti
        sscabiei — Sarcoptes scabiei
        tadhaerens — Trichoplax adhaerens
        tcastaneum — Tribolium castaneum
        tkitauei — Thelohanellus kitauei
        tspiralis — Trichinella spiralis
        turticae — Tetranychus urticae
        znevadensis — Zootermopsis nevadensis

    Ensembl Genomes Plants

        alyrata — Arabidopsis lyrata
        atauschii — Aegilops tauschii
        athaliana — Arabidopsis thaliana
        atrichopoda — Amborella trichopoda
        bdistachyon — Brachypodium distachyon
        bnapus — Brassica napus
        boleracea — Brassica oleracea
        brapa — Brassica rapa
        bvulgaris — Beta vulgaris subsp. vulgaris
        ccrispus — Chondrus crispus
        cmerolae — Cyanidioschyzon merolae
        creinhardtii — Chlamydomonas reinhardtii
        gmax — Glycine max
        gsulphuraria — Galdieria sulphuraria
        hvulgare — Hordeum vulgare
        lperrieri — Leersia perrieri
        macuminata — Musa acuminata
        mtruncatula — Medicago truncatula
        obarthii — Oryza barthii
        obrachyantha — Oryza brachyantha
        oglaberrima — Oryza glaberrima
        oglumaepatula — Oryza glumaepatula
        oindica — Oryza sativa indica
        olongistaminata — Oryza longistaminata
        olucimarinus — Ostreococcus lucimarinus
        omeridionalis — Oryza meridionalis
        onivara — Oryza nivara
        opunctata — Oryza punctata
        orufipogon — Oryza rufipogon
        osativa — Oryza sativa Japonica
        ppatens — Physcomitrella patens
        ppersica — Prunus persica
        ptrichocarpa — Populus trichocarpa
        sbicolor — Sorghum bicolor
        sitalica — Setaria italica
        slycopersicum — Solanum lycopersicum
        smoellendorffii — Selaginella moellendorffii
        stuberosum — Solanum tuberosum
        taestivum — Triticum aestivum
        tcacao — Theobroma cacao
        tpratense — Trifolium pratense
        turartu — Triticum urartu
        vvinifera — Vitis vinifera
        zmays — Zea mays
        """)
        sys.exit()

    ##############################################################
    ##############################################################



    ##############################################################
    ##############################################################




    ##############################################################
    ##############################################################
    print('\n')
    if gene_table!=None:
        cell_type_ids, gene_ids, gene_lists = get_gene_lists_from_bool_table(gene_table)
    elif gene_lists!=None:
        cell_type_ids, gene_ids, gene_lists = get_gene_lists_from_files(gene_lists)
    else:
        print("didn't get any values for -gene_table OR -gene_lists. We need at least one of them!")
        sys.exit()

    print(len(gene_ids))

    #################
    ## print out some basic summaries on the input data
    print(len(cell_type_ids),'cell types found')
    print(cell_type_ids[:min([4,len(cell_type_ids)])],'...\n')

    print(len(gene_ids),'genes found')
    print(gene_ids[:min([4,len(gene_ids)])],'...\n')

    ## make a hash table to look up the index from a cell type name
    cell_type_idx_dict = {key:value for value, key in enumerate(cell_type_ids)}
    # for c in cell_type_ids:
    #     print(c, cell_type_idx_dict[c])
    #################

    #################
    ## if we need to convert over to human ids, this is the first thing we'll need to do
    if convert_to_human:
        print('converting to human')
        temp_org_to_human, human_to_temp_org = get_orthologue_dicts(gp, gene_ids, organism, 'hsapiens')
        #get_orthologue_dicts(gene_ids[:10], organism, 'hsapiens')

        ## update the gene_ids and gene_lists to reflect the conversion to human
        gene_ids, gene_lists = update_converted_genes(temp_org_to_human, gene_ids, gene_lists)
        organism = 'hsapiens'
        print('\n\nfinished converting IDs\n\n')

        print('gene IDs now look like:')
        print(gene_ids[:4],'...\n')
        print('enrichment lists now look like:')
        print('\t',cell_type_ids[0])
        print('\t\t',gene_lists[0][:3])
        print('\t',cell_type_ids[1],'...')
        print('\t\t',gene_lists[1][:3],'...\n\n')



    #################
    print('Converting IDs to ENSP format.')
    print('This is needed for cross-referencing the StringDB database later\n')


    # for i in range(0,len(gene_lists)):
    #     l=gene_lists[i]
    #     if "ENSG00000139874" in l:
    #         print('found "ENSG00000139874" in',i)

    gene_ids, gene_lists, gene_symbol_dict = convert_to_ensp(organism, gp, gene_ids, gene_lists)
    gene_symbol_lists = gene_lists_to_symbol_lists(gene_lists, gene_symbol_dict)
    gene_symbol_list_dicts = gene_symbol_lists_to_dicts(gene_symbol_lists)


    # for i in range(0,len(gene_lists)):
    #     l=gene_lists[i]
    #     if "ENSP00000267377" in l:
    #         print('found "ENSP00000267377" in',i)



    print('finished converting to ENSPs\n')
    print('gene IDs now look like:')
    print(gene_ids[:4],'...\n')
    print('enrichment lists now look like:')
    print('\t',cell_type_ids[0])
    print('\t\t',gene_lists[0][:3])
    #print('\t',cell_type_ids[1],'...')
    #print('\t\t',gene_lists[1][:3],'...\n\n')


    symbols = list(gene_symbol_dict.keys())
    print(len(symbols))
    for i in range(0,5):
        print(symbols[i], gene_symbol_dict[symbols[i]])
    #sys.exit()

    #####################################################################################
    #####################################################################################
    ###################   go term filter for subcellular localization   #################
    #####################################################################################
    #####################################################################################

    print('getting subcellular localizations\n')
    receptor_list, secreted_list = get_receptors_and_secreted(organism, gp, gene_ids)


    print('found',len(receptor_list),'plasma membrane')
    print('found',len(secreted_list),'extracellular')

    # print("in receptor list?:","ENSP00000267377" in receptor_list)
    # print("in secreted list?:","ENSP00000267377" in secreted_list)

    ## if there are any extracellular genes in the plasma membrane gene lists, pop them out
    ## we only want loose ligands in the secreted list 
    for r in receptor_list:
        # print(r)
        # print(secreted_list[:10])
        if r in secreted_list:
            print('removing',r,gene_symbol_dict[r],"from secreted, because it's also annotated as being in the plasma membrane")
            secreted_list.pop(secreted_list.index(r))

    # for r in secreted_list:
    #     # print(r)
    #     # print(secreted_list[:10])
    #     if r in receptor_list:
    #         print('removing',r,gene_symbol_dict[r],"from receptor, because it's also annotated as being in the secreted")
    #         receptor_list.pop(receptor_list.index(r))


    print('found',len(receptor_list),'plasma membrane')
    print('found',len(secreted_list),'extracellular')
    receptor_or_secreted_list = list(set(receptor_list + secreted_list))

    print(len(receptor_or_secreted_list), 'receptors and secreted genes found')

    # if "ENSP00000267377" in receptor_or_secreted_list:
    #     print("found ENSP00000267377 in receptor_or_secreted_list")
    # else:
    #     print("Couldn't find ENSP00000267377 in receptor or secreted!")
    #     sys.exit()

    ## set up a dictionary with the info needed about each gene for writing to file later
    gene_detail_dict = {}
    for g in gene_ids:
        if g in receptor_list:
            gene_detail_dict[g]=[g,gene_symbol_dict[g],'plasma_membrane']
        if g in secreted_list:
            gene_detail_dict[g]=[g,gene_symbol_dict[g],'extracellular']

    ## subset the enriched gene lists for each subcellular localization
    receptor_gene_lists, secreted_gene_lists = filter_enriched_lists_for_subcellular_localization(receptor_list, secreted_list, gene_lists)

    ## print out some summaries
    print('plasma membrane:')
    for i in range(0,len(cell_type_ids)):
        temp_list = receptor_gene_lists[i]
        temp_list_len = len(receptor_gene_lists[i])
        print('\t',cell_type_ids[i],temp_list_len)
        if temp_list_len!=0:
            print('\t\t',receptor_gene_lists[i][0:min([temp_list_len,3])],'...')

    print('\nextracellular:')
    for i in range(0,len(cell_type_ids)):
        temp_list = secreted_gene_lists[i]
        temp_list_len = len(secreted_gene_lists[i])
        print('\t',cell_type_ids[i],temp_list_len)
        if temp_list_len!=0:
            print('\t\t',secreted_gene_lists[i][0:min([temp_list_len,3])],'...')


    # for i in range(0,len(secreted_gene_lists)):
    #     l=secreted_gene_lists[i]
    #     if "ENSP00000267377" in l:
    #         print('found "ENSP00000267377" in secreted_gene_lists',i)


    # for i in range(0,len(receptor_gene_lists)):
    #     l=receptor_gene_lists[i]
    #     if "ENSP00000267377" in l:
    #         print('found "ENSP00000267377" in receptor_gene_lists',i)

    #####################################################################################
    #####################################################################################
    ###################   cross reference all of the receptor and secreted   ############
    ###################       lists with the StringDB action database        ############
    #####################################################################################
    #####################################################################################

    if organism in organism_action_files:
        db_file = stringdb_dir+organism_action_files[organism]

    #########
    ## first we'll load into memory only the interactions that stand
    ## a chance of being pertinent to our dataset
    print('\n\nloading in the interaction table')
    action_table, activation_dict, inhibition_dict, other_dict, all_included_bg = subset_action_table(receptor_or_secreted_list,db_file)
    print(len(action_table)-1,'interactions among the genes remaining')

    # for i in range(0,len(action_table)):
    #     if "ENSP00000267377" in action_table[i]:
    #         print("found ENSP00000267377 in action_table:",action_table[i])


    #########
    ## convert that to an action dictionary so that we can look up the details quickly later
    # first make a unique key for the interaction lists
    action_dict = {}
    symbol_interaction_dict = {}
    for line in action_table:
        temp_id = line[0]+":"+line[1]
        #print(temp_id)
    #    if line[0] in gene_symbol_dict and line[1] in gene_symbol_dict:
        temp_symbol_id = gene_symbol_dict[line[0]]+":"+gene_symbol_dict[line[1]]
    # else:
    #     temp_symbol_id = None
        # print(temp_id,temp_symbol_id)
        if temp_symbol_id not in symbol_interaction_dict:
            symbol_interaction_dict[temp_symbol_id]=True
        if temp_id not in action_dict:
            action_dict[temp_id] = [line]
        else:
            temp_list = action_dict[temp_id]
            temp_list.append(line)
            action_dict[temp_id] = temp_list




    interaction_lists = get_empty_interaction_lists(gene_lists)



    interaction_title_line = [["autocrine_paracrine",
    "cell_type_1","gene_1","gene_1_symbol","gene_1_location",
    "cell_type_2","gene_2","gene_2_symbol","gene_2_location",
    "gene1","gene2","action","inhibition","directional","is_direction","score"]]
    out_summary_table = [["cell_type_1","cell_type_2","count"]]
    out_table = interaction_title_line[:]


    ## make the cell type enrichment dicts
    gene_list_dict = []
    for i in range(0,len(gene_lists)):
        temp_gene_dict = {j:i for i, j in enumerate(gene_lists[i])}
        gene_list_dict.append(temp_gene_dict)



    ## go through the action_dict & find what cell types are enriched for each of the genes
    cell_type_interaction_count = np.zeros((len(cell_type_ids),len(cell_type_ids)))
    all_interactions = list(action_dict.keys())

    # for interaction in all_interactions:
    #     if "ENSP00000267377" in interaction:
    #         print("found ENSP00000267377 in all_interactions:",interaction)

    print('cateloguing which cell types have the interactions')
    for interaction_idx in range(0,len(all_interactions)):
        if interaction_idx %500 == 0:
            print('\t',interaction_idx,'\t',interaction_idx/len(all_interactions))
        interaction = all_interactions[interaction_idx]
        genes = interaction.split(':')
        gene1 = genes[0]
        gene2 = genes[1]
        gene1_symbol = gene_symbol_dict[gene1]
        gene2_symbol = gene_symbol_dict[gene2]
        ## go through all of the cell types and find which cell types are enriched for the genes
        gene1_cell_types = get_cell_types_gene_enriched_in(gene_list_dict, gene_lists, cell_type_ids, gene1)
        gene2_cell_types = get_cell_types_gene_enriched_in(gene_list_dict, gene_lists, cell_type_ids, gene2)



        ## write the output for each of the combinations of cell types
        for cell_type_1 in gene1_cell_types:
            for cell_type_2 in gene2_cell_types:
                if cell_type_1 == cell_type_2:
                    out_line =['autocrine']
                else:
                    out_line = ['paracrine']
                cell_type_1_idx = cell_type_idx_dict[cell_type_1]
                cell_type_2_idx = cell_type_idx_dict[cell_type_2]
                interaction_lists[cell_type_1_idx][cell_type_2_idx].append(gene1_symbol)
                #if gene2_symbol not in interaction_lists[i][j]:
                interaction_lists[cell_type_1_idx][cell_type_2_idx].append(gene2_symbol)
                out_line += [cell_type_1]+gene_detail_dict[gene1]
                out_line += [cell_type_2]+gene_detail_dict[gene2]
                for line in action_dict[gene1+":"+gene2]:
                    out_table.append(out_line+line)

                cell_type_interaction_count[cell_type_1_idx,cell_type_2_idx]+=1

    for i in range(0,len(cell_type_ids)):
        cell_type_1 = cell_type_ids[i]
        for j in range(i,len(cell_type_ids)):
            cell_type_2 = cell_type_ids[j]

            temp_count = cell_type_interaction_count[i,j]

            out_summary_table.append([cell_type_1,cell_type_2,temp_count])
    ## will need to catalogue the number remaining
    #out_summary_table.append([cell_type_1,cell_type_2,temp_count])



    if False:
        ## go through all the receptor receptor pairs
        for i in range(0,len(gene_lists)):
            gene_list_1 = gene_lists[i]
            cell_type_1 = cell_type_ids[i]
            print(cell_type_1)
            for j in range(i,len(gene_lists)):
                cell_type_2 = cell_type_ids[j]
                gene_list_2 = gene_lists[j]
                print('\t',cell_type_2)
                temp_count = 0
                for gene1 in gene_list_1:
                    gene1_symbol = gene_symbol_dict[gene1]
                    for gene2 in gene_list_2:
                        gene2_symbol = gene_symbol_dict[gene2]
                        out_line = None

                        g1_g2=True
                        g2_g1=True
                        try:
                            action_dict[gene1+":"+gene2][0]
                        except:
                            g1_g2=False
                        try:
                            action_dict[gene2+":"+gene1][0]
                        except:
                            g2_g1=False


                        if g1_g2:#ene1+":"+gene2 in action_dict:
                            #if action_dict[gene1+":"+gene2]!=[]:
                                
                            #if gene1_symbol not in interaction_lists[i][j]:
                            interaction_lists[i][j].append(gene1_symbol)
                            #if gene2_symbol not in interaction_lists[i][j]:
                            interaction_lists[i][j].append(gene2_symbol)
                            if cell_type_1 == cell_type_2:
                                out_line = ['autocrine']
                            else:
                                out_line = ['paracrine']
                            out_line += [cell_type_1]+gene_detail_dict[gene1]
                            out_line += [cell_type_2]+gene_detail_dict[gene2]
                            temp_count+=1
                            for line in action_dict[gene1+":"+gene2]:
                                out_table.append(out_line+line)
                                                    
                        if g2_g1:#ene2+":"+gene1 in action_dict:
                            #if action_dict[gene2+":"+gene1]!=[]:
                            #if gene1_symbol not in interaction_lists[i][j]:
                            interaction_lists[i][j].append(gene1_symbol)
                            #if gene2_symbol not in interaction_lists[i][j]:
                            interaction_lists[i][j].append(gene2_symbol)
                            if cell_type_1 == cell_type_2:
                                out_line = ['autocrine']
                            else:
                                out_line = ['paracrine']
                            out_line += [cell_type_2]+gene_detail_dict[gene2]
                            out_line += [cell_type_1]+gene_detail_dict[gene1]
                            temp_count+=1
                            for line in action_dict[gene2+":"+gene1]:
                                out_table.append(out_line+line)
                                #print(out_table[-1])
                                
                print('\t\t',temp_count)
                out_summary_table.append([cell_type_1,cell_type_2,temp_count])

    ## annotate the action table for the mechanism of interaction
    out_table = annotate_interactions_for_activation_and_inhibition(out_table, activation_dict, inhibition_dict, other_dict)

    ## remove the duplicates in the interaction_lists:
    for i in range(0,len(interaction_lists)):
        for j in range(0,len(interaction_lists[i])):
            interaction_lists[i][j]=list(set(interaction_lists[i][j]))


    out_summary_table_copy = out_summary_table[:]
    out_table_copy = out_table[:]



    if not os.path.isdir(out_dir):
        process_dir(out_dir)

    write_table(out_table_copy,out_dir+'/all_cell_type_specific_interactions.tsv')
    write_table(out_summary_table_copy,out_dir+'/all_cell_cell_interaction_summary.tsv')



    output_gprofiler = get_gprofiler_on_interaction_lists(organism, 
                                                          gp, 
                                                          cell_type_ids, 
                                                          interaction_lists, 
                                                          bg = all_included_bg)
    output_gprofiler = annotate_gprofiler_output(symbol_interaction_dict, 
                                                 gene_symbol_lists, 
                                                 cell_type_idx_dict,
                                                 gene_symbol_list_dicts,
                                                 output_gprofiler)
    write_table(output_gprofiler,out_dir+'/all_cell_type_specific_interactions_gprofiler.tsv')
    ##########################################################




    ##########################################################
    ## extracellular:extracellular interactions

    #extracellular_extracellular_lists = get_empty_interaction_lists()
    #receptor_gene_lists, secreted_gene_lists

    # ## go through the out_table and filter it for extracellular:extracellular interactions
    # temp_file_str = out_dir+'/extracellular_extracellular_cell_type_specific_interactions.txt'
    # rm(temp_file_str)
    # extra_cell_file = open(temp_file_str,'a')
    # extra_cell_file.write('\t'.join(interaction_title_line[:])+'\n')
    # #extracellular_extracellular_interactions = interaction_title_line[:]
    # for i in range(1,len(out_table)):
    #     if out_table[i][4]  == 'extracellular' and out_table[i][8] == 'extracellular':
    #         temp_line = out_table[i][:]
    #         temp_line = '\t'.join(temp_line)
    #         if i!=len(out_table)-1:
    #             temp_line+='\n'
    #         extra_cell_file.write(temp_line)
    #         #extracellular_extracellular_interactions.append(out_table[i])
    # # extracellular_extracellular_interactions_copy = extracellular_extracellular_interactions[:]
    # # write_table(extracellular_extracellular_interactions_copy,out_dir+'/extracellular_extracellular_cell_type_specific_interactions.txt')

    print('\n\nwriting output files\n')

    print('extracellular','extracellular')
    temp_file_str = out_dir+'/extracellular_extracellular_cell_type_specific_interactions.txt'
    get_subset(out_table, interaction_title_line, temp_file_str,'extracellular','extracellular')
    print('plasma_membrane','plasma_membrane')
    temp_file_str = out_dir+'/plasma_membrane_plasma_membrane_cell_type_specific_interactions.txt'
    get_subset(out_table, interaction_title_line, temp_file_str,'plasma_membrane','plasma_membrane')

    ##########################################################
    ## extracellular:plasma membrane interactions
    print('extracellular','plasma_membrane')

    prev_interact_dict = {}
    extracellular_pm_interactions = interaction_title_line[:]
    for i in range(1,len(out_table)):
        inuqe_id_1,inuqe_id_2 = get_unique_ids(out_table[i])
        if not key_in_dict(inuqe_id_1,prev_interact_dict) and not (key_in_dict(inuqe_id_2,prev_interact_dict)):# temp_line not in extracellular_pm_interactions:
            if out_table[i][4]  == 'extracellular' and out_table[i][8] == 'plasma_membrane':
                extracellular_pm_interactions.append(out_table[i])
                prev_interact_dict[inuqe_id_1]=True
                prev_interact_dict[inuqe_id_2]=True
            elif out_table[i][4]  == 'plasma_membrane' and out_table[i][8] == 'extracellular':
                ## make it a directional file so that things only flow from extracellular to plasma membrane
                temp_line = out_table[i][:]
                temp_cell_description = temp_line[1:5]
                temp_line[1:5] = temp_line[5:9]
                temp_line[5:9] = temp_cell_description
                #if not key_in_dict(inuqe_id_1,prev_interact_dict) and not (key_in_dict(inuqe_id_2,prev_interact_dict)):# temp_line not in extracellular_pm_interactions:
                extracellular_pm_interactions.append(temp_line)
                prev_interact_dict[inuqe_id_1]=True
                prev_interact_dict[inuqe_id_2]=True


    extracellular_pm_interactions_copy = extracellular_pm_interactions[:]
    write_table(extracellular_pm_interactions_copy,out_dir+'/extracellular_plasma_membrane_cell_type_specific_interactions.txt')

    # ##########################################################
    # ## plasma membrane:plasma membrane interactions

    # pm_pm_interactions = interaction_title_line[:]
    # for i in range(1,len(out_table)):
    #     if out_table[i][4]  == 'plasma_membrane' and out_table[i][8] == 'plasma_membrane':
    #         pm_pm_interactions.append(out_table[i])


    # pm_pm_interactions_copy = pm_pm_interactions[:]
    # write_table(pm_pm_interactions_copy,out_dir+'/plasma_membrane_plasma_membrane_cell_type_specific_interactions.txt')
    return()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-gene_lists",
        help='a comma separated list of files that contain the enriched genes for each cell type.',
        type = str)

    parser.add_argument(
        "-gene_table",'-bool_table','-in_table', '-i',
        help='A boolean True False table indicating whether a gene should be included in each cell types list.',
        type = str)

    parser.add_argument(
        "-organism",'-species','-s',
        help='which organism are we using. (Use the -species_codes to print out a list of all supported species and their codes).',
        type = str,
        default = 'hsapiens')

    parser.add_argument(
        "-species_codes",
        help='print out a list of all supported species and their codes.',
        action = 'store_true',
        default = False)

    parser.add_argument(
        "-convert_to_human",
        help='if using a non-mouse/human species, it may be useful to convert to human, as these interactions will be more complete. Note that you will still have to provide an -organism argument so that we know what species to convert from.',
        action = 'store_true',
        default = False)

    parser.add_argument(
        "-out_dir",'-out','-o',
        help='The output directory to use. If it does not exist, it will be created.',
        type = str)


    parser.add_argument(
        "-stringdb_dir",'-sdb',
        help='The directory containing the StringDB action files',
        type = str,
        default = '/usr/local/lib/cell_signals/')

    args = parser.parse_args()
    pyminer_cell_signals_main(gene_lists = args.gene_lists,
                              gene_table = args.gene_table,
                              out_dir = args.out_dir,
                              organism = args.organism,
                              species_codes = args.species_codes,
                              convert_to_human = args.convert_to_human,
                              stringdb_dir = args.stringdb_dir)





        if args.processes == None:
            args.processes = multiprocessing.cpu_count()
        threads = args.processes
        indices_list = get_contiguous_indices(threads, full_expression.shape[0])
        
        ray.init()

        ###########
        if not args.hdf5:
            ray_full_expression = ray.put(full_expression)
        else:
            ## make copies of the input hdf5 file
            hdf5_file_list = []
            for t in range(threads):
                hdf5_file_list.append(args.infile+"_"+str(t))
                cp(args.infile+" "+hdf5_file_list[-1])
        ###########

        r_jobs = []
        for t in range(threads):
            if not args.hdf5:
                r_jobs.append(ray_get_global_cell_type_aov_stats.remote(indices_list[t], 
                                                                        ray_full_expression,
                                                                        list_of_k_sample_indices))
            else:
                r_jobs.append(ray_get_global_cell_type_aov_stats.remote(indices_list[t], 
                                                                        None,
                                                                        list_of_k_sample_indices,
                                                                        hdf5_file = hdf5_file_list[t]))
        temp_r_results = ray.get(r_jobs)
        ray.shutdown()
        ########################################
    
        global_percent_express_dict_list = []
        global_non_zero_mean_dict_list = []
        k_group_percent_express_dict_list = []
        non_zero_mean_expression_dict_list = []
        all_aov_results = []

        for t in range(len(temp_r_results)):
            global_percent_express_dict_list.append(temp_r_results[t][0])
            global_non_zero_mean_dict_list.append(temp_r_results[t][1])
            k_group_percent_express_dict_list.append(temp_r_results[t][2])
            non_zero_mean_expression_dict_list.append(temp_r_results[t][3])
            all_aov_results.append(temp_r_results[t][4])

        global_percent_express = ray_dicts_to_array(global_percent_express_dict_list)
        global_non_zero_mean = ray_dicts_to_array(global_non_zero_mean_dict_list)
        k_group_percent_express = ray_dicts_to_array(k_group_percent_express_dict_list)
        non_zero_mean_expression = ray_dicts_to_array(non_zero_mean_expression_dict_list)
        all_aov_results = ray_dicts_to_array(temp_aov_results).tolist()

        global_percent_express = global_percent_express[:,0]
        global_non_zero_mean = global_non_zero_mean[:,0]

        print(global_percent_express)
        print(global_non_zero_mean)
        print(k_group_percent_express)
        print(non_zero_mean_expression)
        print(all_aov_results)


def pyminer_get_stats_main():


if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	## global arguments
	parser.add_argument(
	    '-infile','-in','-i','-input',
	    dest='infile',
	    type=str)
	parser.add_argument(
	    "-sample_groups",
	    help='if you know how which samples belong to which groups, feed in a file that has the samples in the first column, and their group number (index starting at 0), in the second column. The IDs must be in the same order as in the infile too.',
	    dest = 'sample_groups',
	    type = str)
	parser.add_argument(
	    "-classes",
	    help='if there are classes to compare, put the annotation table in this argument',
	    dest = 'class_file',
	    type = str)
	parser.add_argument(
	    "-within_group",
	    help='if you know how which samples belong to which groups, feed in a file that has the samples in the first column, and their group number (index starting at 0), in the second column. The IDs must be in the same order as in the infile too.',
	    dest = 'within_group',
	    type = int)
	parser.add_argument(
	    "-out_dir","-o","-out",
	    help='if you know how which samples belong to which groups, feed in a file that has the samples in the first column, and their group number (index starting at 0), in the second column. The IDs must be in the same order as in the infile too.',
	    dest = 'out_dir',
	    type = str)
	parser.add_argument(
	    "-species","-s",
	    help = 'what species is this? Must be gProfiler compatible.',
	    dest = 'species',
	    type = str,
	    default = 'hsapiens')
	parser.add_argument(
	    '-no_gProfile',
	    help = 'should we do the automated gprofiler results?',
	    default = False,
	    action = 'store_true')
	parser.add_argument(
	    "-FDR","-fdr","-fdr_cutoff",
	    help='The desired Benjamini-Hochberg False Discovery Rate (FDR) for multiple comparisons correction (default = 0.05)',
	    dest = 'FDR_cutoff',
	    type = float,
	    default = 0.05)
	parser.add_argument(
	    "-Zscore","-Z_score_cutoff","-Z","-zscore","-z",
	    help='The desired False Discovery Rate (FDR) for multiple comparisons correction (default = 0.05)',
	    dest = 'Zscore',
	    type = float,
	    default = 2.0)
	parser.add_argument(
	    '-hdf5',
	    help = 'The input file is an HDF5 file',
	    default = False,
	    action = 'store_true')
	parser.add_argument(
	    "-ID_list","-ids",
	    help = 'If we are using an hdf5 file, give the row-wise IDs in this new line delimeted file',
	    type = str)
	parser.add_argument(
	    "-columns","-cols",
	    help = 'If we are using an hdf5 file, give the column-wise IDs in this new line delimeted file',
	    type = str)
	parser.add_argument(
	    '-rows',
	    help = 'if the samples are in rows, and variables are in columns',
	    default = False,
	    action = 'store_true')
	parser.add_argument(
	    "-log",'-log2','-log_transform',
	    help='do a log transformation prior to clustering',
	    action = 'store_true',
	    default = False)
	parser.add_argument(
	    '-lin_norm',
	    help = 'should we normalize the rows before doing the stats?',
	    default = False,
	    action = 'store_true')
	parser.add_argument(
	    '-processes', '-p',
	    help = 'The number of processes to use. Default will be the number of available threads.',
	    default = None,
	    type = int)
	args = parser.parse_args()


	pyminer_get_stats_main(,
		                   ,
		                   ,
		                   ,
		                   ,
		                   ,
		                   ,
		                   ,
		                   ,
		                   ,
		                   ,)


#!/usr/bin/env python3
##import dependency libraries
import sys,time,glob,os,pickle,fileinput, random, h5py
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
from statsmodels.nonparametric.smoothers_lowess import lowess
from sklearn.neighbors import RadiusNeighborsRegressor as neighbors
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
import argparse

try:
    from pyminer.common_functions import *
except:
    from common_functions import *

##############################################################

def dispersion_main(args):
    if args.out_dir[-1]!='/':
    	args.out_dir+='/'
    if not os.path.isdir(args.out_dir):
        process_dir(args.out_dir)
    ##############################################################
    gc.enable()
    ## read in the input file
    if not args.hdf5:
        in_mat_str = np.array(read_table(args.infile))
        in_mat = np.array(in_mat_str[1:,1:],dtype=float)
        ID_list = in_mat_str[1:,0].tolist()
        del in_mat_str
    else:
        ## read in the hdf5 file
        print('reading hdf5 file:\n\t',args.infile)
        h5f = h5py.File(args.infile, 'r')
        in_mat=h5f["infile"]
        ID_list = read_file(args.ids,'lines')




    if args.log:
        if np.min(in_mat)<0.0:
            print("can't log transform with negative values... Bringing the matrix up to min = 1")
            in_mat+=min(in_mat)
        in_mat = np.log2(1 + in_mat)



    ## calculate the means
    means = np.mean(in_mat,axis=1)
    ## calculate the coefficient of variance
    variance = np.std(in_mat,axis=1)#/(means+1)
    ## square the coefficient of variance
    variance*=variance

    ## remove the nans

    print(np.sum(np.isnan(variance)))
    variance[np.isnan(variance)]=0
    print(np.sum(np.isnan(variance)))


    linear_dim = np.shape(means)


    ## plot (a random sample of) the relationship between the coefficient of variance and means
    max_sample_size = 50000
    sample_size = min([max_sample_size,linear_dim[0]])
    if sample_size < max_sample_size:
    	print('sampling the distribution for',sample_size,'points')
    	index_vect  = np.arange(linear_dim[0])
    	np.random.shuffle(index_vect)
    	sample = index_vect[:sample_size].tolist()
    	## always include the min and max values for interpolation
    	x_max_index = np.where(np.array(means) == np.max(means))[0][0]
    	x_min_index = np.where(np.array(means) == np.min(means))[0][0]
    	#print(sample)
    	#print(type(sample))
    	#sample = sample.tolist()
    	if x_max_index not in sample:
    		sample.append(x_max_index)
    	if x_min_index not in sample:
    		sample.append(x_min_index)
    	#print(sample)
    	print(type(sample))
    else:
        sample = np.arange(linear_dim[0], dtype =  int).tolist()
        #print(sample)
        #sys.exit()

    sample = np.sort(sample)
    ## remove the indices where the variance is zero
    keep_sample = np.where(variance[sample]>0)[0]
    sample = sample[keep_sample]

    plt.scatter(means[sample], variance[sample], c = 'black', s = 0.5)#,xlab = 'mean',ylab = 'CV')
    #plt.show()
    #sys.exit()

    ## do the lowess fit. This returns the expected values of the variance
    print('calculating the best fit curve')

    ## 

    neigh = neighbors(radius = 1.0, weights = 'uniform', leaf_size = 30)
    mean_max = max(means)
    mean_min = min(means)
    sds_min = min(variance)

    ## first calculate the lowess curve on the sample
    lowess_estimates_sample = lowess(variance[sample]-sds_min+1,means[sample]-mean_min+1, delta = 0.01*mean_max)-1
    # lowess_estimates_sample[:,0] = lowess_estimates_sample[:,0] + mean_min
    # lowess_estimates_sample[:,1] = lowess_estimates_sample[:,1] + sds_min

    ## remove the nans 
    lowess_estimates_sample[np.isnan(lowess_estimates_sample)]=0

    print(lowess_estimates_sample)

    ## remove x ties from the lowess as this messes up the interpolation
    values, indices = np.unique(lowess_estimates_sample[:,0], return_index=True)
    lowess_estimates_sample = np.array(lowess_estimates_sample[indices,:])
    print(lowess_estimates_sample)
    print(type(lowess_estimates_sample))
    #sys.exit()


    # ## positive control
    # x = np.linspace(0, 10, num=11, endpoint=True)
    # print(x)
    # y = np.cos(-x**2/9.0)
    # print(y)
    # f2 = interp1d(x, y, kind='cubic')
    # print(f2)
    # xnew = np.linspace(0, 10, num=41, endpoint=True)
    # print(xnew)
    # print(f2(xnew))

    print(np.shape(lowess_estimates_sample[:,0]))
    print(np.shape(lowess_estimates_sample[:,1]))

    ## cross the spread
    print("original_min_max:",min(lowess_estimates_sample[:,0]),max(lowess_estimates_sample[:,0]))


    epsilon = (max(lowess_estimates_sample[:,0]) - min(lowess_estimates_sample[:,0]))*0.01
    print('epsilon:',epsilon)

    #make the new x values evenly spaced a
    temp_min = min(lowess_estimates_sample[:,0])# - epsilon
    temp_max = max(lowess_estimates_sample[:,0])# + epsilon
    new_x = np.linspace(temp_min,temp_max, num=sample_size, endpoint = True)
    #new_x = np.linspace(temp_min,temp_max, num=sample_size, endpoint = True)
    #new_x[0] = temp_min
    #new_x[-1] = temp_max

    interpolation = interp1d(lowess_estimates_sample[:,0], lowess_estimates_sample[:,1], kind='cubic')


    print("new_min_max:",min(new_x),max(new_x))
    print(new_x)

    ## interpolate the lowess function
    new_y = interpolation(new_x)
    print(new_x,new_y)

    ## get the interplated lowess ready for nearest neighbor regression
    train_x = np.array([[x] for x in lowess_estimates_sample[:,0].tolist()])
    train_y = lowess_estimates_sample[:,1]

    full_x = np.array([[x] for x in means.tolist()])
    #train_x = full_x[sample]
    #train_y = variance[sample]

    print(train_x)
    print(train_y)
    #neigh.fit(train_x,train_y)
    print(new_x)
    print(new_y)
    new_x = np.array([[x] for x in new_x.tolist()])

    #plt.clf()
    plt.scatter(new_x,new_y, c = 'red', s = 0.75)
    plt.savefig(args.out_dir+'local_fit.png',dpi=600,bbox_inches='tight')
    #plt.show()
    #sys.exit()
    neigh.fit(new_x,new_y)

    ## now calculate the values for everything else
    bin_size = 100000
    total_vars = len(means)
    bins = []
    cur_bin = 0
    while cur_bin<total_vars:
    	bins.append(min(cur_bin, total_vars))
    	cur_bin+=bin_size

    bins.append(total_vars)
    print(bins)

    lowess_estimates = np.zeros((len(means),))
    # for i in range(1,len(bins)):
    # 	print("working on",bins[i-1],bins[i])
    # 	# if i%50000 == 0:
    # 	# 	print('\t',bins[i]/len(means))
    # 	lowess_estimates[bins[i-1]:bins[i]] += neigh.predict(full_x[bins[i-1]:bins[i]])

    for i in range(0,total_vars):
        x_dif_vect = np.abs(full_x[i]-new_x)
        closest_idx = int(np.argmin(x_dif_vect))
        lowess_estimates[i]=new_y[closest_idx]
        



    #lowess_estimates = neigh.predict(full_x)

    #lowess_estimates = lowess(variance[sample]+1,means[sample]+1)-1#, return_sorted = False)#, frac = 1e-3, delta = 1e-10*max(means))
    #lowess_estimates = [x[0] for x in lowess_estimates]


    #plt.scatter(means[sample],lowess_estimates[sample],c='red',s=0.5)
    #plt.savefig(args.out_dir+'local_fit_sample.png',dpi=600,bbox_inches='tight')
    #plt.show()
    #sys.exit()
    ## calculate the residuals
    print('getting the residuals')
    residuals = variance - lowess_estimates


    # print(lowess_estimates)
    # # sys.exit()
    # print(residuals - variance)
    # sys.exit()
    print('sample_means',means[:10])
    print('sample_estimates',lowess_estimates[:10])
    print('sample_CVs',variance[:10])
    print('sample_residuals',residuals[:10])


    ## reshape it back into the table
    resid_table = residuals

    print(np.sum(np.isnan(residuals)))
    nan_idx = np.where(np.isnan(residuals))[0]

    print('nan means',means[nan_idx])
    print('nan sds',variance[nan_idx])
    print('nan estimates',lowess_estimates[nan_idx])
    print('nan resid',residuals[nan_idx])


    #print(np.nan_std(residuals))
    st_dev_resid = np.std(residuals)
    print('sd residuals:',st_dev_resid)

    st_dev_cutoff = st_dev_resid*args.z_cutoff
    print('residual cutoff for overdispersion:',st_dev_cutoff)

    ## plot the residuals
    plt.clf()
    plt.scatter(means[sample],residuals[sample], c = 'black', s = 0.5)#,xlab = 'mean',ylab = 'CV')
    plt.plot([min(means),max(means)],[0,0],c='red')
    ## find the significant residuals
    resid_sample = residuals[sample]
    mean_sample = means[sample]
    sig_indices = np.where(resid_sample >= st_dev_cutoff)[0]
    plt.scatter(mean_sample[sig_indices], resid_sample[sig_indices], c = 'blue', s = 2)
    plt.savefig(args.out_dir+'residuals.png',dpi=600,bbox_inches='tight')
    #plt.show()


    ## determine which genes are locally overdispersed
    print('calculating local overdispersion')
    local_overdispersion_bool_table = resid_table >= st_dev_cutoff

    print('number of overdispersed genes:')
    number_overdispersed_per_group = np.sum(local_overdispersion_bool_table, axis = 0)
    print(number_overdispersed_per_group)



    ## get the boolean overdispersed table ready for writing to file
    #print(np.shape(local_overdispersion_bool_table))
    #print(local_overdispersion_bool_table)
    local_overdispersion_bool_table = np.array(local_overdispersion_bool_table,dtype = str)
    #print(local_overdispersion_bool_table)
    local_overdispersion_bool_table=local_overdispersion_bool_table.tolist()
    print(local_overdispersion_bool_table[:5],len(local_overdispersion_bool_table))
    print(np.shape(means)[0])
    ## add the gene names
    gene_names = ID_list
    for i in range(0,(np.shape(means)[0])):
        #print(i)
        local_overdispersion_bool_table[i] = [gene_names[i]] + [local_overdispersion_bool_table[i]]


    local_overdispersion_bool_table=np.array(local_overdispersion_bool_table, dtype = str)
    print(local_overdispersion_bool_table)
    print(np.shape(local_overdispersion_bool_table))

    write_table(local_overdispersion_bool_table,args.out_dir+'/locally_overdispersed_boolean_table.txt')




if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-infile","-i", 
        help="the input expression matrix",
        type = str)
    parser.add_argument("-out_dir", '-out','-o',
        help="the directory for output files",
        type = str)
    parser.add_argument("-log", 
        help="log transform the means",
        action = 'store_true',
        default = False)
    parser.add_argument("-hdf5", 
        help="the input file is an hdf5 file",
        action = 'store_true',
        default = False)
    parser.add_argument("-ids", '-ID_list','-id_list',
        help="if an hdf5 file is used, we'll need the list of IDs",
        type = str)
    parser.add_argument("-z_cutoff", "-z",
        help="the cutoff for how many standard deviations over the mean residual should be considered overdispersed. Default = 0.5",
        type = float,
        default = .5)
    args = parser.parse_args()
    dispersion_main(args)


#!/usr/bin/env python3

##import dependency libraries
import sys,time,glob,os,pickle,fileinput
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
import argparse
from copy import deepcopy
import pandas as pd
try:
    from pyminer.common_functions import *
except:
    from common_functions import *
##############################################################


##################################################################
## processing functions


def get_group(i,
              sig_q_and_dist, 
              group_names, 
              k_group_max, 
              k_group_means):
    if np.sum(k_group_means[i]) == 0.0:
        return(["None","None"])
    indexes = np.where(k_group_means[i] == k_group_max[i])[0]
    ## I can't believe this actually happened, but I came accross a dataset with a 
    ## non-zero tie for mean expression
    if indexes.size > 1:
        return(["None","None"])
    else:
        index = indexes[0]
    if not sig_q_and_dist[i]:
        return(["None", group_names[index]])
    else:
        return([group_names[index], group_names[index]])


def process_symbol(in_symbol_list,ensg):
    if len(in_symbol_list)==0:
        return(ensg)
    else:
        return('_'.join(in_symbol_list))


def process_def(in_def):
    if len(in_def)==0:
        return("NA")
    else:
        return(in_def[0])


def get_best_match_for_all_subset(summary_table,
                                  group_names,
                                  top_n_genes = 5):
    headers = summary_table[0]
    summary_table = summary_table[1:]
    print(headers)
    df = pd.DataFrame(summary_table, columns = headers)
    print(df.head())
    previously_selected_genes = []
    selected_genes = {}
    selected_genes_linear = []
    selected_genes_table = []
    #['gene','gene_symbol','gene_def', 'aov_BH_p', 'range', 'dist_from_max_to_second_max', 'q_value', 'specific_marker_for_group','best_match_group']
    df = df.sort_values(by="dist_from_max_to_second_max", ascending=False)
    for group_name in group_names:
        temp_genes = list(df.loc[df["specific_marker_for_group"] == group_name,"gene"])
        len_temp_genes = len(temp_genes)
        capture = min([top_n_genes,len_temp_genes])
        temp_genes = temp_genes[:capture]
        num_genes_left = top_n_genes - len(temp_genes)
        if num_genes_left>0:
            ## in case their aren't a ton of very specific
            ## we'll top it off with genes that are just highly enriched, but not specific
            temp_genes_2 = list(df.loc[df["best_match_group"] == group_name,"gene"])
            capture_2 = min([top_n_genes,num_genes_left])
            temp_genes += list(temp_genes_2[:capture_2])
        selected_genes[group_name] = temp_genes
        selected_genes_linear += temp_genes
        for temp_gene in temp_genes:
            selected_genes_table.append([temp_gene, temp_gene + "_" + group_name])
    return(selected_genes_table)


## </processing functions>
##########################################################################


def get_marker_genes(symbol_def_dict, 
                     means,
                     sig,
                     out,
                     sig_cutoff = 0.05,
                     q_cutoff = 0.5,
                     percentile = 0.9):
    ## import the annotation dict
    symbol_dict, def_dict = import_dict(symbol_def_dict)

    k_group_means_file = np.array(read_table(means))
    id_list = k_group_means_file[1:,0].tolist()
    group_names = k_group_means_file[0,1:].tolist()

    if len(group_names)<3:
    	sys.exit('need at least 3 groups')

    ###########################
    ## get the means array
    k_group_means = np.array(k_group_means_file[1:,1:],dtype=float)
    k_group_min = np.min(k_group_means,axis=1)
    k_group_max = np.max(k_group_means,axis=1)
    k_group_range = k_group_max - k_group_min
    ## get the distance from max to second max
    dist_max_to_second = np.zeros((len(id_list)))
    q_value_vect = np.zeros((len(id_list)))
    for i in range(0,len(id_list)):
    	temp_sorted = np.sort(k_group_means[i,:])[::-1]
    	dist_max_to_second[i]=temp_sorted[0]-temp_sorted[1]
    	q_value_vect[i] = dist_max_to_second[i]/k_group_range[i]

    sorted_dist_to_max = np.sort(dist_max_to_second[:],kind="mergesort")
    index_for_cutoff = int(len(id_list)*percentile)
    dist_max_to_second_cutoff = sorted_dist_to_max[index_for_cutoff]

    ##########################
    ## 
    BH_corrected_aov_file = np.array(read_table(sig))
    BH_corrected_aov = np.array(BH_corrected_aov_file[1:,3],dtype=float)

    ##
    ## only consider significant genes
    sig_bool = BH_corrected_aov<sig_cutoff
    q_bool = q_value_vect>q_cutoff
    dist_to_second_max_bool = dist_max_to_second>dist_max_to_second_cutoff
    print('\n\n\nsig_bool',sig_bool,sum(sig_bool))
    print('q_bool',q_bool,sum(q_bool))
    print('dist_to_second_max_bool',dist_to_second_max_bool,sum(dist_to_second_max_bool))
    sig_q_and_dist = (np.array(sig_bool,dtype=int) + np.array(q_bool,dtype=int) + np.array(dist_to_second_max_bool,dtype=int))==3
    print('\ncombined\n',sig_q_and_dist)


    print('found',sum(sig_q_and_dist),'highly expressed, somewhat "exclusive" marker genes')

    print('making the output table...')

    out_file = [['gene','gene_symbol','gene_def', 'aov_BH_p', 'range', 'dist_from_max_to_second_max', 'q_value', 'specific_marker_for_group','best_match_group']]

    for i in range(0,len(id_list)):
        ## in the case of entrez IDs
        try:
            int(float(id_list[i]))
        except:
            temp_id = id_list[i]
        else:
            temp_id = str(int(float(id_list[i])))
        temp_line = []
        temp_line.append(temp_id)
        temp_line.append(process_symbol(symbol_dict[temp_id],temp_id))
        temp_line.append(process_def(def_dict[temp_id]))
        temp_line.append(BH_corrected_aov[i])
        temp_line.append(k_group_range[i])
        temp_line.append(dist_max_to_second[i])
        temp_line.append(q_value_vect[i])
        temp_line+=get_group(i,
                             sig_q_and_dist, 
                             group_names, 
                             k_group_max, 
                             k_group_means)
        out_file.append(temp_line)

    ## get the input for plot_subset...
    plot_subset_input = []
    for i in range(0,len(out_file)):
    	if out_file[i][-1]!="None":
    		plot_subset_input.append([out_file[i][0],out_file[i][1]])

    best_match_for_all_subset = get_best_match_for_all_subset(deepcopy(out_file),
                                                              group_names)

    subset_best_for_all_markers_file = os.path.join(out,"best_sorted_markers.tsv")
    subset_input_file = os.path.join(out,"subset_input.txt")
    print("out:",out)
    print("subset_input_file",subset_input_file)
    print("subset_best_for_all_markers_file",subset_best_for_all_markers_file)
    write_table(plot_subset_input,subset_input_file)
    write_table(best_match_for_all_subset, subset_best_for_all_markers_file)
    write_table(out_file,os.path.join(out,'marker_gene_annotations.tsv'))
    #cmd("plot_subset.py -i ""-plot_subset "+subset_input_file)
    return()


if __name__ ==  "__main__":
    parser = argparse.ArgumentParser()
    ## global arguments
    parser.add_argument(
        '-means','-m','-mean',
        dest='means',
        type=str)
    parser.add_argument(
        '-significance','-sig','-anovas','-anova', '-aov',
        dest='sig',
        type=str)
    parser.add_argument(
        '-out','-o',
        help = "output directory",
        dest='out',
        type=str)
    parser.add_argument(
        '-annotation_dict','-ad',
        dest='symbol_def_dict',
        type=str)
    args = parser.parse_args()
    get_marker_genes(symbol_def_dict=args.symbol_def_dict, 
                     means=args.means,
                     sig=args.sig,
                     out=args.out)

